import json

import pytest
from flask import Flask
from flask_restalchemy.tests.sample_model import db


@pytest.fixture
def client(flask_app, db_session):
    from werkzeug.wrappers import BaseResponse

    def response_wrapper(*args):
        resp = BaseResponse(*args)
        if resp.status_code in [200, 201]:
            resp.parsed_data = json.loads(resp.data.decode('utf-8'))
        return resp

    test_client = flask_app.test_client()
    test_client.response_wrapper = response_wrapper
    return test_client


@pytest.fixture()
def flask_app():
    app = Flask('flask_sqlapi_sample')
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
    app.config['PROPAGATE_EXCEPTIONS'] = True
    db.init_app(app)
    with app.app_context():
        yield app


@pytest.fixture()
def db_session(flask_app):
    db.create_all()
    yield db.session
    db.session.remove()
    db.drop_all()



from sqlalchemy import DateTime
from sqlalchemy_utils import PasswordType, JSONType

from .modelserializer import ModelSerializer, NestedModelField, PrimaryKeyField, NestedAttributesField, Field

SWAGGER_BASIC_TYPES = {
    str: dict(type='string'),
    bool: dict(type='boolean'),
    int: dict(type='integer', format='int64'),
    float: dict(type='number', format='double'),
    bytes: dict(type='string', format='byte'),
}


def gen_spec(model_serializer: ModelSerializer, http_method: str, is_child=False):
    '''
    Generates a Swagger spec dict for the given `model_serializer` and `http_method`.

    By setting this dict to the `specs_dict` attribute of the respective Resource class method, a full Swagger spec
    can be generated automatic for the API:

        StuffResource.get.specs_dict = gen_spec(stuff_serializer, "GET")

    :param class model_serializer: Resource serializer instance

    :param str http_method: http method used

    :param bool is_child: Tels if the model is a child resource in the route, meaning that parent
        id must be part of the route

    :rtype: dict
    '''
    resource_name = model_serializer.get_model_name()
    parameters = []
    if is_child:
        parameters.append({
            'name': 'parent_id',
            'in': 'path',
            'required': True,
            'type': 'integer',
            'format': 'int64'
        })
    if http_method in ['GET', 'PUT', 'DELETE']:
        parameters.append({
            'name': 'id',
            'in': 'path',
            'required': True,
            'type': 'integer',
            'format': 'int64'
        })
    if http_method in ['POST', 'PUT']:
        parameters.append({
            'name': 'body',
            'in': 'body',
            'required': True,
            'schema': {
                '$ref': '#/definitions/%s' % resource_name
            }
        })

    produces = ["application/json"]
    receives = ["application/json"]

    specs_dict = {
        "tags": [resource_name],
        "parameters": parameters,
        "produces": produces,
        "receives": receives,
        "schemes": [
            "http",
            "https"
        ],
        "deprecated": False,
        "security": [
            {
                "Bearer": []
            }
        ],
        "responses": _gen_path_responses(http_method, resource_name),
        "definitions": _gen_object_definition(model_serializer),
    }

    return specs_dict


def _gen_path_responses(http_method, resource_name):
    '''
    This function receives a HTTP method and a resource name and generates the standard responses of the method

    :param str http_method:
        http method used

    :param str resource_name:
        name of the resource being served

    :rtype dict:
    :return: dict containing the standard responses of the method
    '''
    if http_method == 'GET_Collection':
        return {
            '200': {
                'description': 'successful operation',
                'schema': {
                    'type': 'array',
                    'items': {
                        '$ref': '#/definitions/%s' % resource_name
                    }
                }
            }
        }
    elif http_method == 'PUT':
        return {
            '204': { 'description': 'successfully updated' },
            '400': { 'description': 'invalid %s supplied' % resource_name },
            '404': { 'description': '%s not found' % resource_name }
        }
    elif http_method == 'GET':
        return {
            '200': {
                'description': 'successful operation',
                'schema': {
                    '$ref': '#/definitions/%s' % resource_name
                }
            },
            '400': { 'description': 'invalid ID supplied' },
            '404': { 'description': '%s not found' % resource_name }
        }
    elif http_method == 'POST':
        return {
            '201': { 'description': 'the %s was created successfully' % resource_name },
            '405': { 'description': 'invalid input' }
        }
    elif http_method == 'DELETE':
        return {
            '204': { 'description': 'successfully deleted' },
            '400': { 'description': 'invalid ID supplied' },
            '404': { 'description': '%s not found' % resource_name }
        }

    return None


def _gen_object_definition(model_serializer: ModelSerializer):
    definitions = {}
    properties = {}
    for field_name, field in model_serializer._fields.items():
        if field_name == 'id':
            continue
        elif isinstance(field, NestedModelField):
            nested_resource_name = field.serializer.get_model_name()
            properties[field_name] = {
                "$ref": "#/definitions/{}".format(nested_resource_name)
            }
            definitions.update(_gen_object_definition(field.serializer))
        elif isinstance(field, NestedAttributesField):
            properties[field_name] = {
                'type': 'object',
                'readOnly': True,
                'properties': {}
                }
            for nested_attribute in field.serializer.attributes:
                attr_type = field.serializer.attributes[nested_attribute]
                nested_properties = _gen_object_parameters_from_column(attr_type)
                properties[field_name]['properties'][nested_attribute] = nested_properties
        elif isinstance(field, PrimaryKeyField):
            properties[field_name] = {
                'type': 'array',
                'items': {"type": "integer"}
            }
        elif isinstance(field, Field):
            col = model_serializer.model_columns.get(field_name)
            if col is None:
                continue
            properties[field_name] = _gen_object_parameters_from_column(col.type)
            if field.load_only:
                properties[field_name]['writeOnly'] = True
            if field.dump_only:
                properties[field_name]['readOnly'] = True
    resource_name = model_serializer.get_model_name()
    definitions[resource_name]  = {
        "type": "object",
        "properties": properties
    }
    return definitions


def _gen_object_parameters_from_column(sql_type):
    if isinstance(sql_type, DateTime) or (hasattr(sql_type, "impl") and isinstance(sql_type.impl, DateTime)):
        return {
            'type': 'string',
            'format': 'date-time'
        }
    elif isinstance(sql_type, PasswordType):
        return {
            'type': 'string',
            'format': 'password'
        }
    elif isinstance(sql_type, JSONType):
        return {
            'type': 'string',
        }
    elif isinstance(sql_type, type):
        return dict(SWAGGER_BASIC_TYPES[sql_type])
    elif hasattr(sql_type, 'python_type'):
        return dict(SWAGGER_BASIC_TYPES[sql_type.python_type])
    else:
        return {}



from .api import Api
from .serialization.modelserializer import ModelSerializer, DateTimeSerializer, Field, \
    NestedAttributesField, NestedModelField, PrimaryKeyField
from .serialization.serializer import Serializer



from .serialization.modelserializer import ModelSerializer
from .serialization.swagger_spec import gen_spec


def item_resource_factory(item_resource_class, serializer: ModelSerializer, item_decorators = None):
    '''
    Creates a new Resource class in runtime to allow use of Flasgger to automatic generate Swagger specs,
    since Flaggers needs a different resource class for each REST resource.

    Note: this is inefficient and makes the code harder to read. See #26
    '''

    item_decorators = item_decorators or []

    class _ItemResource(item_resource_class):
        method_decorators = item_decorators

        def get(self, **kw): return item_resource_class.get(self, **kw)

        def put(self, **kw): return item_resource_class.put(self, **kw)

        def delete(self, **kw): return item_resource_class.delete(self, **kw)

        get.specs_dict = gen_spec(serializer, "GET")
        put.specs_dict = gen_spec(serializer, "PUT")
        delete.specs_dict = gen_spec(serializer, "DELETE")

    return _ItemResource


def collection_resource_factory(collection_resource_class, serializer: ModelSerializer, collection_decorators=None):
    '''
    Creates a new Resource class in runtime to allow use of Flasgger to automatic generate Swagger specs,
    since Flaggers needs a different resource class for each REST resource.

    Note: this is inefficient and makes the code harder to read. See #26
    '''

    class _CollectionResource(collection_resource_class):
        method_decorators = collection_decorators

        def get(self, **kw): return collection_resource_class.get(self, **kw)

        def post(self, **kw): return collection_resource_class.post(self, **kw)

        get.specs_dict = gen_spec(serializer, "GET_Collection")
        post.specs_dict = gen_spec(serializer, "POST")


    return _CollectionResource



from typing import Union

from sqlalchemy.orm.dynamic import AppenderMixin
from sqlalchemy.sql import sqltypes

from flask_restalchemy.serialization.serializer import Serializer, DateTimeSerializer, is_datetime_field, is_enum_field, \
    EnumSerializer


class ModelSerializer(Serializer):
    """
    Serializer for SQLAlchemy Declarative classes
    """

    DEFAULT_SERIALIZERS = [
        (is_datetime_field, DateTimeSerializer),
        (is_enum_field, EnumSerializer)
    ]

    def __init__(self, model_class):
        """
        :param Type[DeclarativeMeta] model_class: the SQLAlchemy mapping class to be serialized
        """
        self._mapper_class = model_class
        self._fields = self._get_declared_fields()
        # Collect columns not declared in the serializer
        for column in self.model_columns.keys():
            field = self._fields.setdefault(column, Field())
            # Set a serializer for fields that can not be serialized by default
            if field.serializer is None:
                serializer = self._get_default_serializer(self.model_columns.get(column))
                if serializer:
                    field._serializer = serializer

    def get_model_name(self) -> str:
        return self._mapper_class.__name__

    @property
    def model_columns(self) -> dict:
        return self._mapper_class.__mapper__.c

    def dump(self, model):
        serial = {}
        for attr, field in self._fields.items():
            if field is None:
                serial[attr] = getattr(model, attr)
                continue
            if field.load_only:
                continue
            value = getattr(model, attr) if hasattr(model, attr) else None
            if value is None:
                serialized = None
            elif field.serializer:
                serialized = field.serializer.dump(value)
            else:
                serialized = value
            serial[attr] = serialized
        return serial

    def load(self, serialized, existing_model=None):
        """
        Instancialize a Declarative model from a serialized dict

        :param dict serialized: the serialized object.

        :param None|DeclarativeMeta existing_model: If given, the model will be updated with the serialized data.

        :rtype: DeclarativeMeta
        """
        if existing_model:
            model = existing_model
            # If deserialization must update an existing model, check if primary key is the same
            pk_name = self._get_pk_name(existing_model)
            assert getattr(existing_model, pk_name) == serialized[pk_name], \
                "Primary key value of serialized nested object is inconsistent"
        else:
            model = self._mapper_class()
        for field_name, value in serialized.items():
            field = self._fields[field_name]
            if field.dump_only:
                continue
            if value is None:
                deserial_value = value
            elif field.serializer:
                if isinstance(field.serializer, ModelSerializer):
                    if existing_model:
                        existing_nested = getattr(existing_model, field_name)
                    else:
                        existing_nested = self._find_existing_model(field, value)
                    deserial_value = field.serializer.load(value, existing_nested)
                else:
                    deserial_value = field.serializer.load(value)
            else:
                deserial_value = value
            setattr(model, field_name, deserial_value)
        return model

    def before_put_commit(self, model, session):
        pass

    def after_put_commit(self, model, session):
        pass

    def before_post_commit(self, model, session):
        pass

    def after_post_commit(self, model, session):
        pass

    def _get_default_serializer(self, column):
        for check_type, serializer_class in self.DEFAULT_SERIALIZERS:
            if check_type(column):
                return serializer_class(column)

    @classmethod
    def _get_declared_fields(cls) -> dict:
        fields = {}
        # Fields should be only defined ModelSerializer subclasses,
        if cls is ModelSerializer:
            return fields
        for attr_name in dir(cls):
            if attr_name.startswith('_'):
                continue
            value = getattr(cls, attr_name)
            if isinstance(value, Field):
                fields[attr_name] = value
        return fields

    def _find_existing_model(self, field, value):
        class_mapper = field.serializer._mapper_class
        pk_name = self._get_pk_name(class_mapper)
        pk = value.get(pk_name)
        if pk:
            return class_mapper.query.get(pk)
        else:
            return None

    def _get_pk_name(self, existing_model):
        primary_keys = existing_model.__mapper__.primary_key
        assert len(primary_keys) == 1, "Nested object must have exactly one primary key"
        pk_name = primary_keys[0].key
        return pk_name


class Field(object):
    """
    Configure a ModelSerializer field
    """

    def __init__(self, dump_only=False, load_only=False, serializer=None):
        self.dump_only = dump_only
        self.load_only = load_only
        self._serializer = serializer

    @property
    def serializer(self):
        return self._serializer


class NestedModelField(Field):
    """
    A field to Dump and Update nested models.
    """

    def __init__(self, declarative_class, **kw):
        super().__init__(**kw)
        if self._serializer is None:
            self._serializer = ModelSerializer(declarative_class)


class NestedAttributesField(Field):
    """
    A read-only field that dump nested object attributes.
    """

    class NestedAttributesSerializer(Serializer):

        def __init__(self, attributes, many):
            self.attributes = attributes
            self.many = many

        def dump(self, value):
            if self.many:
                serialized = [self._dump_item(item) for item in value]
            else:
                return self._dump_item(value)
            return serialized

        def _dump_item(self, item):
            serialized = {}
            for attr_name in self.attributes:
                serialized[attr_name] = getattr(item, attr_name)
            return serialized

        def load(self, serialized):
            raise NotImplementedError()


    def __init__(self, attributes: Union[tuple, dict], many=False, **kw):
        super().__init__(serializer=self.NestedAttributesSerializer(attributes, many), **kw)
        assert 'serializer' not in kw, "NestedAttributesField does not support custom Serializer"


class PrimaryKeyField(Field):
    """
    Convert relationships in a list of primary keys (for serialization and deserialization).
    """

    class PrimaryKeySerializer(Serializer):

        def __init__(self, declarative_class):
            self.declarative_class = declarative_class
            self._pk_column = get_model_pk(self.declarative_class)

        def load(self, serialized):
            pk_column = self._pk_column
            return self.declarative_class.query.filter(pk_column.in_(serialized)).all()

        def dump(self, value):
            pk_column = self._pk_column
            if is_tomany_attribute(value):
                serialized = [getattr(item, pk_column.key) for item in value]
            else:
                return getattr(value, pk_column.key)
            return serialized

    def __init__(self, declarative_class, **kw):
        super().__init__(serializer=self.PrimaryKeySerializer(declarative_class), **kw)


def get_model_pk(declarative_class):
    """
    Get the primary key Column object from a Declarative model class

    :param Type[DeclarativeMeta] declarative_class: a Declarative class

    :rtype: Column
    """
    primary_keys = declarative_class.__mapper__.primary_key
    assert len(primary_keys) == 1, "Nested object must have exactly one primary key"
    return primary_keys[0]


def is_tomany_attribute(value):
    """
    Check if the Declarative relationship attribute represents a to-many relationship.

    :param value: a SQLAlchemy Declarative class relationship attribute

    :rtype: bool
    """
    return isinstance(value, (list, AppenderMixin))



from sqlalchemy import desc, or_, and_
import json
import operator

# Filter operators defined on SqlAlchemy ColumnElement
SQLA_OPERATORS = {
    'like': 'like',
    'notlike': 'notlike',
    'ilike': 'ilike',
    'notilike': 'notilike',
    'is': 'is_',
    'isnot': 'isnot',
    'match': 'match',
    'startswith': 'startswith',
    'endswith': 'endswith',
    'contains': 'contains',
    'in': 'in_',
    'notin': 'notin_',
    'between': 'between',
    'eq': '__eq__',
    'ne': '__ne__',
    'gt': '__gt__',
    'ge': '__ge__',
    'lt': '__lt__',
    'le': '__le__',
}


def parse_value(value, serializer):
    if not serializer:
        return value
    if isinstance(value, list):
        return [serializer.load(item) for item in value]
    return serializer.load(value)


def get_operator(column, op_name, value, serializer):
    """
    :param column:
         SQLAlchemy ColumnElement

    :param op_name:
        Key of OPERATORS or COMPARE_OPERATORS

    :param value:
        value to be applied to the operator

    :rtype: ColumnOperators
    :return:
        returns a boolean, comparison, and other operators for ColumnElement expressions.
        ref: http://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.operators.ColumnOperators
    """
    if not op_name:
        return column.operate(operator.eq, parse_value(value, serializer))
    elif op_name in SQLA_OPERATORS:
        op = SQLA_OPERATORS.get(op_name)
        if op == 'between':
            return column.between(parse_value(value[0], serializer), parse_value(value[1], serializer))
        return getattr(column, op)(parse_value(value, serializer))
    else:
        raise ValueError('Unknown operator {}'.format(op_name))


def get_field_serializer_or_none(serializer, field_name):
    field = serializer._fields.get(field_name)
    if not field:
        return None
    return field.serializer


def apply_request_args_to_query(query, model, model_serializer, request):
    """
        Build a query using query parameters in the http URL, disposed on the request args.
        The default logical operator is AND, but you can set the OR as in the following examples:

            a) OR -> ?filter={"$or":{"name": {"startswith": "Terrans 1"},"location": "Location 1"}}
            b) AND -> ?filter={"$and":{"name": {"ilike": "%Terrans 1%"},"location": "Location 1"}}
                or ?filter={"name": {"ilike": "%Terrans 1%"},"location": {"eq": "Location 1"}}

        Ordered search is available using 'order_by=<col_name>'. The minus sign ("-<col_name>") could be
        used to set descending order.

        :param query:
            SQLAlchemy query instance

        :param class model:
            SQLAlchemy model class representing a database resource

        :param model_serializer:
            instance of model serializer

        :param request:
            Flask http request data

        :rtype: query
        :return: SQLAlchemy query instance
        """

    def build_filter_operator(column_name, request_filter, serializer):
        if column_name == '$or':
            return or_(build_filter_operator(attr, value, serializer) for attr, value in request_filter.items())
        elif column_name == '$and':
            return and_(build_filter_operator(attr, value, serializer) for attr, value in request_filter.items())
        if isinstance(request_filter, dict):
            op_name = next(iter(request_filter))
            return get_operator(getattr(model, column_name), op_name, request_filter.get(op_name),
                                get_field_serializer_or_none(serializer, column_name))
        return get_operator(getattr(model, column_name), None, request_filter,
                            get_field_serializer_or_none(serializer, column_name))

    if 'filter' in request.args:
        filters = json.loads(request.args['filter'])
        for attr, value in filters.items():
            query = query.filter(build_filter_operator(attr, value, model_serializer))
    if 'order_by' in request.args:
        fields = request.args['order_by'].split(',')
        for field in fields:
            if field[0] == '-':
                col = getattr(model, field[1:])
                query = query.order_by(desc(col))
            else:
                col = getattr(model, field)
                query = query.order_by(col)
    # limit and pagination have to be done after order_by
    if 'limit' in request.args:
        limit = request.args['limit']
        query = query.limit(limit)

    return query


def query_from_request(model, model_serializer, request, query=None):
    """
    Perform a filtered search in the database model table using query parameters in the http URL,
    disposed on the request args.

    See function apply_request_args_to_query on how to use filters
    py:function:: apply_request_args_to_query(query, model, model_serializer, request)

    It also paginate the response if a 'page' value is present in the query parameters. A 'per_page'
    value in the query parameters defines the page length, default to 20 items.

    :param class model:
        SQLAlchemy model class representing a database resource

    :param model_serializer:
        instance of model serializer

    :param request:
        Flask http request data

    :param query:
        SQLAlchemy query instance

    :rtype: list|dict
    :return: the serialized response: "if 'page' is defined in the query params, a dict with page, per page, count and results is returned,
    otherwise returns a list of serialized objects"
    """
    if not query:
        query = model.query

    query = apply_request_args_to_query(query, model, model_serializer, request)

    if 'page' in request.args:
        data = query.paginate()
        return {
            'page': data.page,
            'per_page': data.per_page,
            'count': data.total,
            'results': [model_serializer.dump(item) for item in data.items]
        }
    else:
        data = query.all()
        return [model_serializer.dump(item) for item in data]



from abc import ABC, abstractmethod
from datetime import datetime, timezone, timedelta
from enum import Enum

from sqlalchemy import DateTime
import re


class Serializer(ABC):

    def __init__(self, column):
        self.column = column

    @abstractmethod
    def dump(self, value): pass

    @abstractmethod
    def load(self, serialized): pass


class DateTimeSerializer(Serializer):
    """
    Serializer for DateTime objects
    """

    DATETIME_REGEX = "(?P<Y>\d{2,4})-(?P<m>\d{2})-(?P<d>\d{2})" + \
                     "[T ]" + \
                     "(?P<H>\d{2}):(?P<M>\d{2})(:(?P<S>\d{2}))?(\.(?P<f>\d+))?" + \
                     "(?P<tz>([\+-]\d{2}:?\d{2})|[Zz])?"

    DATETIME_RE = re.compile(DATETIME_REGEX)


    def dump(self, value):
        return value.isoformat()

    def load(self, serialized):
        match = self.DATETIME_RE.match(serialized)
        if not match:
            raise ValueError("Could not parse DateTime: '{}'".format(serialized))
        parts = match.groupdict()
        dt = datetime(
            int(parts["Y"]), int(parts["m"]), int(parts["d"]),
            int(parts["H"]), int(parts["M"]), int(parts.get("S") or 0), int(parts.get("f") or 0),
            tzinfo=self._parse_tzinfo(parts["tz"])
        )
        return dt

    def _parse_tzinfo(self, offset_str):
        if not offset_str:
            return None
        elif offset_str.upper() == 'Z':
            return timezone.utc
        else:
            hours = int(offset_str[:3])
            minutes = int(offset_str[-2:])
            # Invert minutes sign if hours == 0
            if offset_str[0] == "-" and hours == 0:
                minutes = -minutes
            return timezone(timedelta(hours=hours, minutes=minutes))


def is_datetime_field(col):
    """
    Check if a column is DateTime (or implements DateTime)

    :param Column col: the column object to be checked

    :rtype: bool
    """
    if hasattr(col.type, "impl"):
        return type(col.type.impl) is DateTime
    else:
        return type(col.type) is DateTime


class EnumSerializer(Serializer):

    def dump(self, value):
        if not value:
            return None
        return value.value

    def load(self, serialized):
        enum = getattr(self.column.type, 'enum_class')
        return enum(serialized)

def is_enum_field(col):
    return hasattr(col.type, 'enum_class') and getattr(col.type, 'enum_class')



from setuptools import setup, find_packages

install_requires=[
    'flask-restful>=0.3.0',
    'flask-sqlalchemy >= 2.3.0',
    'flask',
    'sqlalchemy>=1.1',
]

setup(
    name='flask-restalchemy',
    version='0.11.0',
    packages=find_packages(exclude=['*.tests']),
    url='https://github.com/ESSS/flask-restalchemy',
    license='MIT',
    author='ESSS',
    author_email='dev@esss.co',
    description='Flask extension to build REST APIs based on SQLAlchemy models ',
    keywords='flask sqlalchemy orm',
    data_files=[("", ["LICENSE"])],
    install_requires=install_requires,
    python_requires='>=3.4',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database :: Front-Ends',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



from flask_restful import Api as RestfulApi

from flask_restalchemy.resourcefactory import item_resource_factory, collection_resource_factory
from flask_restalchemy.resources import CollectionRelationResource, ItemRelationResource, \
    CollectionPropertyResource, ItemResource, CollectionResource


class Api(object):

    def __init__(self, app_or_bp=None, prefix='', errors=None, request_decorators=None):
        """
        :param (Flask|Blueprint) app_or_bp: Flask application or Blueprint

        :param str prefix: API endpoints prefix

        :param errors: A dictionary to define a custom response for each
            exception or error raised during a request

        :param callable request_decorators: request decorators for this API object (see
            Flask-Restful decorators docs for more information)
        """
        self.restful_api = RestfulApi(app=app_or_bp, prefix=prefix, decorators=request_decorators,
                                      default_mediatype='application/json', errors=errors)
        if app_or_bp:
            self.init_app(app_or_bp)
        self._db = None

    def init_app(self, app):
        self.restful_api.init_app(app)

    def add_model(self, model, url=None, serializer_class=None, request_decorators=None,
                  collection_decorators=None, collection_name=None):
        """
        Create API endpoints for the given SQLAlchemy declarative class.


        :param class model: the SQLAlchemy declarative class

        :param string url: one or more url routes to match for the resource, standard
             flask routing rules apply. Defaults to model name in lower case.

        :param string collection_name: custom name for the collection endpoint url definition, if not set the model
            table name will be used

        :param Type[ModelSerializer] serializer_class: If `None`, a default serializer will be created.

        :param list|dict request_decorators: decorators to be applied to HTTP methods. Could be a list of decorators
            or a dict mapping HTTP method types to a list of decorators (dict keys should be 'get', 'post' or 'put').
            See https://flask-restful.readthedocs.io/en/latest/extending.html#resource-method-decorators for more
            details.

        :param list|dict collection_decorators: decorators to be applied to HTTP methods for collections. It defaults to
            request_decorators value.
        """
        restful = self.restful_api
        collection_name = collection_name or model.__tablename__
        if not serializer_class:
            serializer = self.create_default_serializer(model)
        else:
            serializer = serializer_class(model)
        url = url or '/' + collection_name.lower()

        if not request_decorators:
            request_decorators = []
        if not collection_decorators:
            collection_decorators = request_decorators

        _ItemResource = item_resource_factory(ItemResource, serializer, request_decorators)
        _CollectionResource = collection_resource_factory(CollectionResource, serializer, collection_decorators)

        restful.add_resource(
            _CollectionResource,
            url,
            endpoint=collection_name + '-list',
            resource_class_args=(model, serializer, self.get_db_session),
        )
        restful.add_resource(
            _ItemResource,
            url + '/<id>',
            endpoint=collection_name,
            resource_class_args=(model, serializer, self.get_db_session)
        )

    def add_relation(self, relation_property, url_rule=None, serializer_class=None,
                     request_decorators=None,
                     collection_decorators=None, endpoint_name=None):
        """
        Create API endpoints for the given SQLAlchemy relationship.

        :param relation_property: model relationship representing the collection to receive the CRUD operations

        :param string url_rule: one or more url routes to match for the resource, standard
             flask routing rules apply. Defaults to model name in lower case.

        :param Type[ModelSerializer] serializer_class: If `None`, a default serializer will be created.

        :param list|dict request_decorators: decorators to be applied to HTTP methods. Could be a list of decorators
            or a dict mapping HTTP method types to a list of decorators (dict keys should be 'get', 'post' or 'put').
            See https://flask-restful.readthedocs.io/en/latest/extending.html#resource-method-decorators for more
            details.

        :param list|dict collection_decorators: decorators to be applied to HTTP methods for collections. It defaults to
            request_decorators value.

        :param string endpoint_name: endpoint name (defaults to :meth:`{model_collection_name}-{related_collection_name}-relation`
            Can be used to reference this route in :class:`fields.Url` fields

        """
        model = relation_property.prop.mapper.class_
        related_model = relation_property.class_
        model_collection_name = model.__tablename__.lower()
        related_collection_name = related_model.__tablename__.lower()
        endpoint_name = endpoint_name or '{}-{}-relation'.format(model_collection_name, related_collection_name)
        if not serializer_class:
            serializer = self.create_default_serializer(model)
        else:
            serializer = serializer_class(model)
        if url_rule:
            assert '<relation_id>' in url_rule
        else:
            url_rule = '/{}/<relation_id>/{}'.format(related_collection_name, relation_property.key)

        if not request_decorators:
            request_decorators = []
        if not collection_decorators:
            collection_decorators = request_decorators

        _ItemRelationResource = item_resource_factory(
            ItemRelationResource,
            serializer,
            request_decorators
        )
        _CollectionRelationResource = collection_resource_factory(
            CollectionRelationResource,
            serializer,
            collection_decorators
        )

        self._add_item_collection_resources(
            _ItemRelationResource,
            _CollectionRelationResource,
            url_rule,
            endpoint_name,
            resource_init_args=(relation_property, serializer, self.get_db_session),
        )

    def _add_item_collection_resources(self, item_resource, collection_resource, url_rule, endpoint,
                                       resource_init_args):
        self.add_resource(
            item_resource,
            url_rule + '/<id>',
            endpoint=endpoint,
            resource_class_args=resource_init_args,
        )
        self.add_resource(
            collection_resource,
            url_rule,
            endpoint=endpoint + '-list',
            resource_class_args=resource_init_args,
        )

    def add_property(self, model, related_model, property_name, url_rule=None,
                     serializer_class=None, request_decorators=[]):
        if not serializer_class:
            serializer = self.create_default_serializer(model)
        else:
            serializer = serializer_class(model)
        related_collection_name = related_model.__tablename__.lower()
        if url_rule:
            assert '<relation_id>' in url_rule
        else:
            url_rule = '/{}/<relation_id>/{}'.format(related_collection_name, property_name.lower())

        class _CollectionPropertyResource(CollectionPropertyResource):
            method_decorators = request_decorators

        self.add_resource(
            _CollectionPropertyResource,
            url_rule,
            endpoint='{}-{}-property'.format(related_collection_name, property_name),
            resource_class_args=(
            model, related_model, property_name, serializer, self.get_db_session)
        )

    def add_resource(self, *args, **kw):
        self.restful_api.add_resource(*args, **kw)

    @staticmethod
    def create_default_serializer(model_class):
        """
        Create a default serializer for the given SQLAlchemy declarative class. Recipe based on
        https://marshmallow-sqlalchemy.readthedocs.io/en/latest/recipes.html#automatically-generating-schemas-for-sqlalchemy-models

        :param model_class: the SQLAlchemy mapped class

        :rtype: class
        """
        from flask_restalchemy.serialization.modelserializer import ModelSerializer
        return ModelSerializer(model_class)

    def get_db_session(self):
        """
        Returns an SQLAlchemy object session. Used by flask-restful Resources to access
        the database.
        """
        if not self._db:
            # Get the Flask application
            if self.restful_api.blueprint_setup:
                flask_app = self.restful_api.blueprint_setup.app
            else:
                flask_app = self.restful_api.app
            assert flask_app and flask_app.extensions, "Flask App not initialized yey"
            self._db = flask_app.extensions['sqlalchemy'].db
        return self._db.session



import warnings

from flask import request, json
from flask_restful import Resource
from sqlalchemy.orm import load_only
from sqlalchemy.orm.collections import InstrumentedList

from flask_restalchemy.serialization.modelserializer import ModelSerializer
from .utils import query_from_request


class BaseResource(Resource):

    def __init__(self, declarative_model, serializer, session_getter):
        """
        The Base class for ORM resources

        :param class declarative_model: the SQLAlchemy declarative class.

        :param ModelSchema serializer: Marshmallow schema for serialization. If `None`, a default serializer will be
            created.

        :param callable session_getter: a callable that returns the DB session. A callable is used since a reference to
            DB session may not be available on the resource initialization.
        """
        self._resource_model = declarative_model
        self._serializer = serializer
        self._serializer.strict = True
        assert isinstance(self._serializer, ModelSerializer), 'Invalid serializer instance: {}'.format(serializer)
        self._session_getter = session_getter

    def save_from_request(self, extra_attrs={}):
        session = self._session_getter()
        model_obj = self._serializer.load(load_request_data())
        for attr_name, value in extra_attrs.items():
            setattr(model_obj, attr_name, value)
        session.add(model_obj)
        session.commit()
        return self._serializer.dump(model_obj).data

    def _save_model(self, model_obj, method):
        session = self._session_getter()
        session.add(model_obj)

        # run pre commit hooks
        if method == 'POST':
            self._serializer.before_post_commit(model_obj, session)
        elif method == 'PUT':
            self._serializer.before_put_commit(model_obj, session)

        session.commit()

        # run post commit hooks
        if method == 'POST':
            self._serializer.after_post_commit(model_obj, session)
        elif method == 'PUT':
            self._serializer.after_put_commit(model_obj, session)

    def _save_serialized(self, serialized_data, existing_model=None):
        model_obj = self._serializer.load(serialized_data, existing_model)

        method = 'PUT' if existing_model else 'POST'
        self._save_model(model_obj, method)

        return self._serializer.dump(model_obj)

    @property
    def _db_session(self):
        return self._session_getter()


class ItemResource(BaseResource):
    """
    flask-restful resource class that receives receives a SQLAlchemy model class and define the API to provide GET,
    UPDATE and DELETE over a single data identified by id
    """

    def get(self, id):
        data = self._resource_model.query.get(id)
        if data is None:
            return NOT_FOUND_ERROR, 404
        return self._serializer.dump(data)

    def put(self, id):
        data = self._resource_model.query.get(id)
        if data is None:
            return NOT_FOUND_ERROR, 404
        serialized = self._serializer.dump(data)
        serialized.update(load_request_data())
        return self._save_serialized(serialized, data)

    def delete(self, id):
        data = self._resource_model.query.get(id)
        if data is None:
            return NOT_FOUND_ERROR, 404
        session = self._db_session
        session.delete(data)
        session.commit()
        return '', 204


class CollectionResource(BaseResource):
    """
    flask-restful resource class that receives a SQLAlchemy model class and define the API to provide LIST and
    CREATE over data of that class
    """

    def get(self):
        data = query_from_request(self._resource_model, self._serializer, request)
        return data

    def post(self):
        saved = self._save_serialized(load_request_data())
        return saved, 201


class CollectionRelationResource(BaseResource):
    """
    flask-restful resource class that receives two SQLAlchemy models, a parent model and a child model,
    and define the API to provide LIST and CREATE over data of the child model associated with a specific
    element of the parent model.
    """

    def __init__(self, relation_property, serializer, session_getter):
        """
        The Base class for ORM resources

        :param class declarative_model: the SQLAlchemy declarative class.

        :param ModelSchema serializer: Marshmallow schema for serialization. If `None`, a default serializer will be
            created.

        :param callable session_getter: a callable that returns the DB session. A callable is used since a reference to
            DB session may not be available on the resource initialization.
        """
        resource_model = relation_property.prop.mapper.class_
        super(CollectionRelationResource, self).__init__(resource_model, serializer, session_getter)
        self._relation_property = relation_property
        self._related_model = relation_property.class_

    def get(self, relation_id):
        session = self._db_session()

        # using options(load_only('id')) avoid unintended subquerying, as all we want is check if the element exists
        related_obj = session.query(self._related_model).options(load_only("id")).get(relation_id)
        if related_obj is None:
            return NOT_FOUND_ERROR, 404

        # TODO: Is there a more efficient way than using getattr?
        relation_list_or_query = getattr(related_obj, self._relation_property.key)
        if isinstance(relation_list_or_query, InstrumentedList) or not hasattr(relation_list_or_query, 'paginate'):
            warnings.warn('Warnning: relationship does not support pagination nor filter. Use flask-sqlalchemy '
                         'relationship with lazy="dynamic".')
            collection = [self._serializer.dump(item) for item in relation_list_or_query]
        else:
            collection = query_from_request(self._resource_model, self._serializer, request, query=relation_list_or_query)
        return collection

    def post(self, relation_id):
        session = self._db_session()
        related_obj = session.query(self._related_model).get(relation_id)
        if not related_obj:
            return NOT_FOUND_ERROR, 404
        collection = getattr(related_obj, self._relation_property.key)
        data_dict = load_request_data()
        resource_id = data_dict.get('id', None)

        if resource_id is not None:
            return self.append_existent(collection, resource_id, session)

        new_obj = self._serializer.load(data_dict)
        session.add(new_obj)
        collection.append(new_obj)

        self._save_model(new_obj, 'POST')
        return self._serializer.dump(new_obj), 201

    def append_existent(self, collection, resource_id, session):
        resource_obj = session.query(self._resource_model).get(resource_id)
        if resource_obj is None:
            return NOT_FOUND_ERROR, 404
        collection.append(resource_obj)
        session.commit()
        return self._serializer.dump(resource_obj), 200


class ItemRelationResource(BaseResource):
    """
    flask-restful resource class that receives two SQLAlchemy models, a parent model and a child model,
    and define the API to provide GET, PUT and DELETE operations over data of the child model associated with a
    specific element of the parent model.
    """

    def __init__(self, relation_property, serializer, session_getter):
        """
        The Base class for ORM resources

        :param class declarative_model: the SQLAlchemy declarative class.

        :param ModelSchema serializer: Marshmallow schema for serialization. If `None`, a default serializer will be
            created.

        :param callable session_getter: a callable that returns the DB session. A callable is used since a reference to
            DB session may not be available on the resource initialization.
        """
        resource_model = relation_property.prop.mapper.class_
        super(ItemRelationResource, self).__init__(resource_model, serializer, session_getter)
        self._relation_property = relation_property
        self._related_model = relation_property.class_

    def get(self, relation_id, id):
        requested_obj = self._query_related_obj(relation_id, id)
        if not requested_obj:
            return NOT_FOUND_ERROR, 404
        return self._serializer.dump(requested_obj), 200

    def put(self, relation_id, id):
        requested_obj = self._query_related_obj(relation_id, id)
        if not requested_obj:
            return NOT_FOUND_ERROR, 404
        serialized = self._serializer.dump(requested_obj)
        serialized.update(load_request_data())
        return self._save_serialized(serialized, requested_obj)

    def delete(self, relation_id, id):
        requested_obj = self._query_related_obj(relation_id, id)
        if not requested_obj:
            return NOT_FOUND_ERROR, 404
        session = self._db_session()
        session.delete(requested_obj)
        session.commit()
        return '', 204

    def _query_related_obj(self, relation_id, id):
        """
        Query resource model by ID but also add the relationship as a query constrain.

        :param relation_id: id of the related model
        :param id: id of the model being required
        :return: model with 'id' that has a related model with 'related_id'
        """

        # This checks if there is a parent with the related child on its relation property
        related = self._db_session.query(self._related_model).options(load_only("id"))\
            .filter(
                self._related_model.id == relation_id,
                self._relation_property.any(id=id)
            ).one_or_none()

        if related is None:
            return None

        return self._db_session.query(self._resource_model).get(id)


class CollectionPropertyResource(CollectionRelationResource):

    def __init__(self, declarative_model, related_model, property_name, serializer, session_getter):
        super(CollectionRelationResource, self).__init__(declarative_model, serializer, session_getter)
        self._related_model = related_model
        self._property_name = property_name

    def get(self, relation_id):
        session = self._db_session()
        related_obj = session.query(self._related_model).get(relation_id)
        if related_obj is None:
            return NOT_FOUND_ERROR, 404
        relation_list_or_query = getattr(related_obj, self._property_name)
        if isinstance(relation_list_or_query, InstrumentedList) or not hasattr(relation_list_or_query, 'paginate'):
            warnings.warn('Warnning: property ' + self._property_name + ' does not support pagination nor filter.'
                          ' Use flask-sqlalchemy and make your property return a query object')
            collection = [self._serializer.dump(item) for item in relation_list_or_query]
        else:
            collection = query_from_request(self._resource_model, self._serializer, request,
                                            query=relation_list_or_query)
        return collection

    def post(self, relation_id):
        return 'POST not allowed for property resources', 405


def load_request_data():
    """
    Returns request data as dict.

    :rtype: dict
    """
    if request.data:
        return json.loads(request.data.decode('utf-8'))
    else:
        return request.form.to_dict()


NOT_FOUND_ERROR = 'Resource not found in the database!'
